/*
 *    This program is free software; you can redistribute it and/or modify
 *    it under the terms of the GNU General Public License as published by
 *    the Free Software Foundation; either version 2 of the License, or
 *    (at your option) any later version.
 *
 *    This program is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *    GNU General Public License for more details.
 *
 *    You should have received a copy of the GNU General Public License
 *    along with this program; if not, write to the Free Software
 *    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 *    main.cpp
 *	  The main file implementing the application of the Transactional
 *    K-Means algorithm
 *    Copyright (C) 2001 Cristian Gozzi,Giuseppe Manco
 *
 *    TO DO:
 *     - improve the I/O Functions 
 */

#include <iostream>
#include <fstream>

#include <cmath>
#include <ctime>
#include "dist.h"

#include "File.h"
#include "AsciiFile.h"
#include "Kmeans.h"
#include "Result.h"
#include "platform/temp_and_unlink.h"


int main(int argc, char** argv) {
	int i, k=0;                   /* local variables                   */
//	SymTab trtable;				  /* Symbol table for transactions	   */	
	char    *s;                   /* to traverse options               */
	char    **optarg = NULL;      /* option argument                   */
	char    *fn_in   = NULL;      /* name of input  file               */
	char    *fn_out  = NULL;      /* name of output file               */
	char    *blanks  = NULL;      /* blanks                            */
	char    *fldseps = NULL;      /* field  separators                 */
	char    *recseps = NULL;      /* record separators                 */
	long    seed      = 0;        /* random seed					   */
	double  supp     = 5.0;       /* minimal support    (in percent)   */
	double  conf     = 10.0;      /* minimal confidence (in percent)   */
	int     clust    = 10;        /* maximal number of clusters        */
	int     maxlen   = 100;       /* maximal number of iterations      */
	int     skip     = 0;         /* number of records to skip         */
	bool    isrec    = 0;         /* flag for one item set per record  */
	bool    load     = 1;         /* flag for loading item sets        */
	bool    XMLFormat   = false;     /* flag for output in XML format       */
	bool    mapping  = false;     /* flag for show mapping of transactions */
	int     maxcnt   = 0;         /* maximal number of items per set   */
	double  maxtrash = 100.0;      /* maximal trash size (in percentage)   */
#ifndef QUIET                     /* if not quiet version,             */
	clock_t t;                  /* timer                             */
#endif
#ifndef QUIET                     /* if not quiet version              */
	char* prgname = argv[0];      /* get program name for error msgs.  */
	/* --- print usage message ---                                     */
	if (argc > 1) {               /* if arguments are given            */
		msg("%s - Clustering mining "
			"with kmeans method\n", argv[0]);
		msg(VERSION);                 /* print a startup message */
	}                             /* print a startup message           */
	else {                        /* if no arguments given             */
		std::cout << "usage: " << argv[0] 
			<< " [options] infile outfile\n";
		std::cout << argv[0] 
			<< " - find clusters of itemsets "
			<< "with kmeans algorithm\n";
		std::cout << VERSION; 
		std::cout << "Options:\n";
		std::cout << "-s#      minimal support of a cluster        "
			<<              "(default: " << supp << ")\n";
		std::cout << "-k#      maximal number of clusters          "
			<<              "(default: " << clust << ")\n";
		std::cout << "-c#      minimal confidence of a cluster     "
			<<              "(default: " << conf << ")\n";
		std::cout << "-n#      maximal number of iterations        "
			<<              "(default: " << maxlen << ")\n";
		std::cout << "-y       do not load item sets into memory   "
			<<              "(work on input file)\n";
		std::cout << "-h#      maximal trash percentage desired    "
			<<              "(default: " << maxtrash << "%)\n";
		std::cout << "-a       show transaction assignements       "
				<<          "(default: false)\n";
		std::cout << "-x       output in XML format                "
				<<          "(default: false)\n";
		std::cout << "-i#      ignore first # records of infile    "
			<<              "(default: " << skip << ")\n";
		std::cout << "-l       one set per record instead of set id and item\n";
		std::cout << "-b/f/r#  blanks, field and record separators "
			<<              "(default: \" \\t\", \" \\t\", \"\\n\")\n";
		std::cout << "infile   file to read item sets from\n";
		std::cout << "outfile  file to write association rules to\n";
		return 0;                   /* print a usage message             */
	}                               /* and abort the program             */
#endif  /* #ifndef QUIET */
	/* --- evaluate arguments -----------------------------------------  */
	for (i = 1; i < argc; i++) {    /* traverse arguments                */
		s = argv[i];                /* get option argument               */
		if (optarg) { *optarg = s; optarg = NULL; continue; }
		if (*s == '-') {            /* -- if argument is an option       */
			if (!*++s) { k++; continue; }
			while (*s) {            /* traverse options                  */
				switch (*s++) {     /* evaluate switches */
				case 's': supp   =      strtod(s, &s);     break;
				case 'c': conf   =      strtod(s, &s);     break;
				case 'k': clust  = (int)strtol(s, &s, 10); break;
				case 'h': maxtrash  =  strtod(s, &s);      break;
				case 'n': maxlen = (int)strtol(s, &s, 10); break;
				case 'i': skip   = (int)strtol(s, &s, 10); break;
				case 'l': isrec  = true;                   break;
				case 'y': load   = false;                  break;
				case 'a': mapping  = true;                 break;
				case 'x': XMLFormat   = true;	               break;
				case 'b': optarg = &blanks;                break;
				case 'f': optarg = &fldseps;               break;
				case 'r': optarg = &recseps;               break;
				default : error(E_OPTION, *--s);           break;
				}                   /* set option variables              */
				if (optarg && *s) { *optarg = s; optarg = NULL; break; }
			} 
		}                           /* get option argument               */
		else {                      /* -- if argument is no option       */
			switch (k++) {          /* evaluate non-options              */
			case  0: fn_in  = s;      break;
			case  1: fn_out = s;      break;
			default: error(E_ARGCNT); break;
			}                       /* note filenames                    */
		}
	}                    /* check arguments                   */
	if ((supp  <  0) || (supp > 100)) error(E_SUPP, supp);
	if ((conf  <  0) || (conf > 100)) error(E_CONF, conf);
	if (maxlen <= 0) error(E_RLEN, maxlen);
	if ((k < 2) || (k > 2))         /* check the number of arguments     */
		error(E_ARGCNT);            /* (only in/out allowed)             */
	if ((!fn_in || !*fn_in) && (k > 2))
		error(E_STDIN);             /* stdin must not be used twice      */
	
	/* ----------------------------------------------------------------- */
	/* ------- Prepare the input/output data --------------------------- */
	/* ----------------------------------------------------------------- */
	File* fp;
	std::ofstream op(fn_out);
//	switch (input_type){
//	case 2:
//		fp = new XMLFile(fn_in,blanks,fldseps,recseps,load,isrec);
//		break;
//	case 1:
//		fp = new SynthetizedFile(fn_in,blanks,fldseps,recseps,load,isrec);
//		break;
//	default:
//		fp = new AsciiFile(fn_in,blanks,fldseps,recseps,load,isrec);
//		break;
//	}
	fp = new AsciiFile(fn_in,blanks,fldseps,recseps,load,isrec);
	

	/* ----------------------------------------------------------------- */
	/* ------- Prepare the dataset ------------------------------------- */
	/* ----------------------------------------------------------------- */
	std::vector<std::string> names;               /* useful to store the filenames */
	char* tmplate = "/temp/tempXXXXXX"; /* temporary filename            */
	int iteration = 0;                  /* the number of iterations      */
	names.push_back(fn_in);             /* store the initial filename    */
	char name[22];                      /* to contain the name           */
	bool trash = false;		            /* check the trash size          */
	do 	{
#ifndef QUIET                                    /* if not quiet version          */
		msg("reading %s ... ", names[iteration].c_str());
		t = clock();                                 /* start the timer           */
#endif                                           /* print a log message           */
		strcpy(name,names[iteration].c_str());
		if (!fp)
			fp = new AsciiFile(name,blanks,fldseps,recseps,load,isrec);
#ifndef QUIET                                    /* if not quiet version          */
		msg("[%d item(s), %d set(s)] done ", fp->items(), fp->size());
		msg("[%.2fs].\n", (clock()-t) /(double)CLOCKS_PER_SEC);
#endif
		int val = fp->size();         /* the current number of itemsets            */
		if (iteration == 0)                     /* compute the maximum allowed    */
			maxtrash = (maxtrash*val)/100;/* trash size                     */

		/* ---------------------------------------------------------------------- */
		/* -------  Main cycle of the algorithm --------------------------------- */
		/* ---------------------------------------------------------------------- */
		Kmeans algo(clust+1,val,fp,XMLFormat);
		/* ---------------------------------------------------------------------- */
		/* ------- Set random cluster seeds ------------------------------------- */
		/* ---------------------------------------------------------------------- */
		algo.setseeds(seed);         /* computes the initial random seeds         */
		/* ---------------------------------------------------------------------- */
		/* ------- Main cycle: compute cluster assigments ----------------------- */
		/* ---------------------------------------------------------------------- */
#ifndef QUIET                              /* if not quiet version                */
		t = clock();                       /* start the timer                     */
#endif
		algo.cycle(conf, maxlen);      /* Computes Cluster assigments             */
		/*  the conf confidence value and the      */
		/*  maxlen maximal number of iterations     */
#ifndef QUIET                          /* if not quiet version                    */
		msg(" done [%.2fs].\n", (clock()-t) /(double)CLOCKS_PER_SEC);
#endif
		/* ---------------------------------------------------------------------- */
		/* ------- Clusters are computed. Print the results --------------------- */
		/* ---------------------------------------------------------------------- */
//		op << endl << algo << endl << flush;/* prepares the output file */
		/* ---------------------------------------------------------------------- */
		/* ------- Manage the trash --------------------------------------------- */
		/* ---------------------------------------------------------------------- */
		strcpy(name,tmplate);                    /* copy the template filename    */
		const char* result = maketempfile(name);
        if (result == 0){    /* generate the temporary file   */
			trash = false;                       /* if unable to generate, exit   */
#ifndef QUIET                                    /* if not quiet version          */
		msg("failed to generate temporary files.\n"); 
#endif                                           /* print a log message           */
		}
		else {                                   /* otherwise, store the trash       */
			names.push_back(result);             /* remember the name (for deletion) */
			std::ofstream tmpin(result); /* create a stream                  */
			Result res(&algo,mapping,XMLFormat);
			res.print(op,iteration);
			op << std::endl;
			algo.dump(tmpin,clust);                 /* and dump the trash to the stream */
			trash = (algo[clust].act_population > maxtrash); /* test whether the trash size matters  */
			if (XMLFormat && !trash)
				op << "</experiment>\n";
			iteration++;                             /* keep track of the current iteration          */             
			clust = 5;                               /* and update the desired number of clusters    */
		}
		delete fp;
		fp = 0;
	}
	while (trash);                               /* iteration continues until trash size matters */
#ifndef QUIET                                    /* if not quiet version          */
		msg("generating output results..."); 
#endif                                           /* print a log message           */
	

#ifndef QUIET                                    /* if not quiet version                         */
		msg("done\n"); 
		msg("Deleting temporary files... ");
#endif                                           /* print a log message                          */
	for (i = 1; i < names.size(); i++)           /* delete the list of temporary files           */
		if( unlinkfile( names[i].c_str() ) == -1 )  
			perror( "Could not delete " );
#ifndef QUIET                                    /* if not quiet version                         */
		msg(" done.\n");
#endif
	return 0;
}

